#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tool to interrogate IDAES flowsheets and list the physical properties
required to simulate it.
"""
# TODO: Missing docstrings
# pylint: disable=missing-function-docstring

import sys
from inspect import isclass

# Import Pyomo libraries
from pyomo.environ import Set, Var, units as pyunits
from pyomo.common.config import ConfigValue

# Import IDAES cores
from idaes.core import (
    declare_process_block_class,
    MaterialFlowBasis,
    PhysicalParameterBlock,
    StateBlockData,
    StateBlock,
    MaterialBalanceType,
    EnergyBalanceType,
    UnitModelBlockData,
    Phase,
    LiquidPhase,
    VaporPhase,
    Component,
)
from idaes.core.util.exceptions import ConfigurationError
import idaes.logger as idaeslog

# Some more information about this module
__author__ = "Andrew Lee"


# Set up logger
_log = idaeslog.getLogger(__name__)


@declare_process_block_class("PropertyInterrogatorBlock")
class PropertyInterrogatorData(PhysicalParameterBlock):
    """
    Interrogator Parameter Block Class

    This class contains the methods and attributes for recording and displaying
    the properties required by the flowsheet.
    """

    CONFIG = PhysicalParameterBlock.CONFIG()

    CONFIG.declare(
        "phase_list",
        ConfigValue(
            domain=dict,
            description="User defined phase list. Dict with form {name: Type}",
        ),
    )
    CONFIG.declare(
        "component_list",
        ConfigValue(
            domain=dict,
            description="User defined component list. Dict with form {name: Type}",
        ),
    )

    def build(self):
        """
        Callable method for Block construction.
        """
        super(PropertyInterrogatorData, self).build()

        self._state_block_class = InterrogatorStateBlock

        # Phase objects
        if self.config.phase_list is None:
            self.Liq = LiquidPhase()
            self.Vap = VaporPhase()
        else:
            for p, t in self.config.phase_list.items():
                if t is None:
                    t = Phase
                elif not isclass(t) or not issubclass(t, Phase):
                    raise ConfigurationError(
                        f"{self.name} invalid phase type {t} (for phase {p})."
                        f" Type must be a subclass of Phase."
                    )
                self.add_component(p, t())

        # Component objects
        if self.config.component_list is None:
            self.A = Component()
            self.B = Component()
        else:
            for j, t in self.config.component_list.items():
                if t is None:
                    t = Component
                elif not isclass(t) or not issubclass(t, Component):
                    raise ConfigurationError(
                        f"{self.name} invalid component type {t} (for "
                        f"component {j}). Type must be a subclass of "
                        f"Component."
                    )
                self.add_component(j, t())

        # Set up dict to record property calls
        self.required_properties = {}

        # Dummy phase equilibrium definition so we can handle flash cases
        self.phase_equilibrium_idx = Set(initialize=[1])

        self.phase_equilibrium_list = {1: ["A", ("Vap", "Liq")]}

    def list_required_properties(self):
        """
        Method to list all thermophysical properties required by the flowsheet.

        Args:
            None

        Returns:
            A list of properties required
        """
        return list(self.required_properties)

    def list_models_requiring_property(self, prop):
        """
        Method to list all models in the flowsheet requiring the given
        property.

        Args:
            prop : the property of interest

        Returns:
            A list of unit model names which require prop
        """
        try:
            return self.required_properties[prop]
        except KeyError:
            raise KeyError(
                "Property {} does not appear in required_properties. "
                "Please check the spelling of the property that you are "
                "interested in.".format(prop)
            )

    def list_properties_required_by_model(self, model):
        """
        Method to list all thermophysical properties required by a given unit
        model.

        Args:
            model : the unit model of interest. Can be given as either a model
                    component or the unit name as a string

        Returns:
            A list of thermophysical properties required by model
        """
        prop_list = []
        if not isinstance(model, str):
            model = model.name

        for k, v in self.required_properties.items():
            if model in v:
                prop_list.append(k)

        if len(prop_list) < 1:
            raise ValueError(
                "Model {} does not appear in the flowsheet. Please check "
                "the spelling of the model provided."
            )
        else:
            return prop_list

    def print_required_properties(self, ostream=None):
        """
        Method to print a summary of the thermophysical properties required by
        the flowsheet.

        Args:
            ostream : output stream to print to. If not provided will print to
                      sys.stdout

        Returns:
            None
        """
        if ostream is None:
            ostream = sys.stdout

        # Write header
        max_str_length = 74
        tab = " " * 4
        ostream.write("\n" + "=" * max_str_length + "\n")
        ostream.write("Property Interrogator Summary" + "\n")
        ostream.write(
            "\n"
            + "The Flowsheet requires the following properties "
            + "(times required):"
            + "\n"
            + "\n"
        )
        for k, v in self.required_properties.items():
            lead_str = tab + k
            trail_str = str(len(v))
            mid_str = " " * (max_str_length - len(lead_str) - len(trail_str))
            ostream.write(lead_str + mid_str + trail_str + "\n")
        ostream.write(
            "\n"
            + "Note: User constraints may require additional properties "
            + "which are not"
            + "\n"
            + "reported here."
            + "\n"
        )

    def print_models_requiring_property(self, prop, ostream=None):
        """
        Method to print a summary of the models in the flowsheet requiring a
        given property.

        Args:
            prop : the property of interest.
            ostream : output stream to print to. If not provided will print to
                      sys.stdout

        Returns:
            None
        """
        if ostream is None:
            ostream = sys.stdout

        tab = " " * 4

        ostream.write("\n")
        ostream.write(
            f"The following models in the Flowsheet " f"require {prop}:" + "\n"
        )

        for m in self.required_properties[prop]:
            ostream.write(tab + m + "\n")

    def print_properties_required_by_model(self, model, ostream=None):
        """
        Method to print a summary of the thermophysical properties required by
        a given unit model.

        Args:
            model : the unit model of interest.
            ostream : output stream to print to. If not provided will print to
                      sys.stdout

        Returns:
            None
        """
        if not isinstance(model, str):
            model = model.name

        if ostream is None:
            ostream = sys.stdout

        tab = " " * 4

        ostream.write("\n")
        ostream.write(
            f"The following properties are required by model " f"{model}:" + "\n"
        )

        for m in self.list_properties_required_by_model(model):
            ostream.write(tab + m + "\n")

    @classmethod
    def define_metadata(cls, obj):
        obj.add_default_units(
            {
                "time": pyunits.s,
                "length": pyunits.m,
                "mass": pyunits.kg,
                "amount": pyunits.mol,
                "temperature": pyunits.K,
            }
        )


class _InterrogatorStateBlock(StateBlock):
    """
    This Class contains methods which should be applied to Property Blocks as a
    whole, rather than individual elements of indexed Property Blocks.
    """

    def initialize(blk, *args, **kwargs):
        """
        Dummy initialization routine, This will raise an TypeError if a user
        tries to initialize a model using the Interrogator Property Package
        and tell them that the model cannot be solved.
        """
        raise TypeError(
            "Models constructed using the Property Interrogator package "
            "cannot be used to solve a flowsheet. Please rebuild your "
            "flowsheet using a valid property package."
        )


@declare_process_block_class(
    "InterrogatorStateBlock", block_class=_InterrogatorStateBlock
)
class InterrogatorStateBlockData(StateBlockData):
    """
    A dummy state block for interrogating flowsheets and recording physical
    properties called for during construction.
    """

    def build(self):
        """
        Callable method for Block construction
        """
        super(InterrogatorStateBlockData, self).build()

        # Add dummy vars for building Ports and returning expressions
        self._dummy_var = Var(initialize=1)
        self._dummy_var_phase = Var(self.params.phase_list, initialize=1)
        self._dummy_var_comp = Var(self.params.component_list, initialize=1)
        self._dummy_var_phase_comp = Var(
            self.params.phase_list, self.params.component_list, initialize=1
        )
        # T and P are often involved in unit conversion checks, so need to have units
        self._dummy_var_T = Var(initialize=1, units=pyunits.K)
        self._dummy_var_P = Var(initialize=1, units=pyunits.Pa)

    # Define standard methods and log calls before returning dummy variable
    def get_material_flow_terms(self, p, j):
        self._log_call("material flow terms")
        return self._dummy_var

    def get_enthalpy_flow_terms(self, p):
        self._log_call("enthalpy flow terms")
        return self._dummy_var

    def get_material_density_terms(self, p, j):
        self._log_call("material density terms")
        return self._dummy_var

    def get_energy_density_terms(self, p):
        self._log_call("energy density terms")
        return self._dummy_var

    # Set default values for required attributes so construction doesn't fail
    def default_material_balance_type(self):
        return MaterialBalanceType.componentPhase

    def default_energy_balance_type(self):
        return EnergyBalanceType.enthalpyTotal

    def define_state_vars(b):
        return {"_dummy_var": b._dummy_var}

    def define_display_vars(b):
        raise TypeError(
            "Models constructed using the Property Interrogator package "
            "should not be used for report methods."
        )

    def get_material_flow_basis(b):
        return MaterialFlowBasis.molar

    def __getattr__(self, prop):
        """
        Overload getattr to log each call for an unknown attribute, assuming
        these are all properties.
        Then, return a dummy variable with the correct indexing set.
        """
        # Log call
        self._log_call(prop)

        # Return dummy var
        if prop.endswith("_phase_comp"):
            return self._dummy_var_phase_comp
        elif prop.endswith("_phase"):
            return self._dummy_var_phase
        elif prop.endswith("_comp"):
            return self._dummy_var_comp
        elif prop == "temperature":
            # Need this for unit conversion checks in some models
            return self._dummy_var_T
        elif prop == "pressure":
            # Need this for unit conversion checks in some models
            return self._dummy_var_P
        else:
            return self._dummy_var

    def _log_call(self, prop):
        """
        Method to log calls for properties in required_properties dict
        """
        # Get the required_properties dict from parameter block
        prop_dict = self.params.required_properties

        # Get name of parent unit to record in required_properties
        name = self._get_parent_unit_name()

        try:
            # If name is not listed for current property, add to list
            if name not in prop_dict[prop]:
                prop_dict[prop].append(name)
        except KeyError:
            # If a KeyError occurs, it means property has not been logged
            # before, so add new entry to dict
            prop_dict[prop] = [name]

    def _get_parent_unit_name(self):
        """
        Method to find the parent unit of the current StateBlock (if one
        exists) and return this so it can be logged as in required_properties.

        If current StateBlock has no parent unit, it is a stand-alone
        StateBlock, so log the name of this instead.
        """
        # Start with current block (i.e. a StateBlock)
        parent = self

        # Search up the parent tree until we find a UnitModel or top of tree
        while True:
            if isinstance(parent, UnitModelBlockData):
                # If parent is a UnitModel, we have found our target
                # Return parent name
                return parent.name
            else:
                if parent.parent_block() is None:
                    # Check if the parent object has no parent, i.e. is top of
                    # tree. If so, we are dealling with a stand-alone
                    # StateBlock.
                    # Return name of parent_component to strip indices
                    return self.parent_component().name
                else:
                    # Otherwise continue searching up tree
                    parent = parent.parent_block()



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for generic reaction package core code

Author: Andrew Lee
"""
import pytest
import re
from sys import modules
from math import log

from pyomo.environ import (
    Block,
    ConcreteModel,
    Constraint,
    Expression,
    exp,
    Set,
    Var,
    value,
    units as pyunits,
)
from pyomo.util.check_units import assert_units_equivalent

from idaes.core.base.property_meta import UnitSet
from idaes.models.properties.modular_properties.base.generic_property import (
    GenericParameterBlock,
)
from idaes.models.properties.modular_properties.base.tests.dummy_eos import DummyEoS

from idaes.models.properties.modular_properties.base.generic_reaction import (
    GenericReactionParameterBlock,
    ConcentrationForm,
)
from idaes.models.properties.modular_properties.reactions.dh_rxn import constant_dh_rxn
from idaes.models.properties.modular_properties.reactions.rate_constant import arrhenius
from idaes.models.properties.modular_properties.reactions.rate_forms import (
    power_law_rate,
)
from idaes.models.properties.modular_properties.reactions.equilibrium_constant import (
    van_t_hoff,
)
from idaes.models.properties.modular_properties.reactions.equilibrium_forms import (
    power_law_equil,
    log_power_law_equil,
)

from idaes.core.util.constants import Constants as constants

from idaes.core.util.exceptions import ConfigurationError, PropertyPackageError
import idaes.logger as idaeslog


# -----------------------------------------------------------------------------
# Dummy methods to avoid errors
def set_metadata(b):
    pass


def define_state(b):
    b.temperature = Var(initialize=100)
    b.mole_frac_phase_comp = Var(
        b.params.phase_list, b.params.component_list, initialize=0.5
    )


# Declare a base units dict to save code later
base_units = {
    "time": pyunits.s,
    "length": pyunits.m,
    "mass": pyunits.kg,
    "amount": pyunits.mol,
    "temperature": pyunits.K,
}


# -----------------------------------------------------------------------------
class TestGenericReactionParameterBlock(object):
    @pytest.fixture
    def m(self):
        m = ConcreteModel()

        # Add a dummy thermo package for validation
        m.params = GenericParameterBlock(
            components={"c1": {}, "c2": {}},
            phases={
                "p1": {"equation_of_state": DummyEoS},
                "p2": {"equation_of_state": DummyEoS},
            },
            state_definition=modules[__name__],
            pressure_ref=100000.0,
            temperature_ref=300,
            base_units=base_units,
        )

        return m

    @pytest.mark.unit
    def test_rate_build(self, m):
        m.rxn_params = GenericReactionParameterBlock(
            property_package=m.params,
            rate_reactions={
                "r1": {
                    "stoichiometry": {("p1", "c1"): -1, ("p1", "c2"): 2},
                    "heat_of_reaction": "foo",
                    "rate_form": "foo",
                }
            },
            base_units=base_units,
        )

        rxn_config = m.rxn_params.config.rate_reactions

        default_units = m.rxn_params.get_metadata().default_units
        assert isinstance(default_units, UnitSet)
        assert_units_equivalent(default_units.TIME, pyunits.s)
        assert_units_equivalent(default_units.LENGTH, pyunits.m)
        assert_units_equivalent(default_units.MASS, pyunits.kg)
        assert_units_equivalent(default_units.AMOUNT, pyunits.mol)
        assert_units_equivalent(default_units.TEMPERATURE, pyunits.K)
        assert_units_equivalent(default_units.CURRENT, pyunits.ampere)
        assert_units_equivalent(default_units.LUMINOUS_INTENSITY, pyunits.candela)

        assert isinstance(m.rxn_params.rate_reaction_idx, Set)
        assert len(m.rxn_params.rate_reaction_idx) == 1
        assert "r1" in m.rxn_params.rate_reaction_idx

        assert not hasattr(self, "equilibrium_reaction_idx")

        assert isinstance(m.rxn_params.rate_reaction_stoichiometry, dict)
        assert len(m.rxn_params.rate_reaction_stoichiometry) == 4
        for k, v in m.rxn_params.rate_reaction_stoichiometry.items():
            if (k[1], k[2]) in rxn_config[k[0]].stoichiometry.keys():
                assert v == rxn_config[k[0]].stoichiometry[k[1], k[2]]
            else:
                assert v == 0

        assert not hasattr(self, "equilibrium_reaction_stoichiometry")

        assert isinstance(m.rxn_params.reaction_idx, Set)
        assert m.rxn_params.reaction_idx == m.rxn_params.rate_reaction_idx

        assert isinstance(m.rxn_params.reaction_r1, Block)

    @pytest.mark.unit
    def test_invalid_unit(self, m):
        with pytest.raises(
            PropertyPackageError,
            match=re.escape(
                "Unrecognized units of measurement for quantity TIME (foo)"
            ),
        ):
            m.rxn_params = GenericReactionParameterBlock(
                property_package=m.params,
                rate_reactions={
                    "r1": {
                        "stoichiometry": {("p1", "c1"): -1, ("p1", "c2"): 2},
                        "heat_of_reaction": "foo",
                        "rate_form": "foo",
                    }
                },
                base_units={
                    "time": "foo",
                    "length": pyunits.m,
                    "mass": pyunits.kg,
                    "amount": pyunits.mol,
                    "temperature": pyunits.K,
                },
            )

    @pytest.mark.unit
    def test_rate_build_no_stoichiometry(self, m):
        with pytest.raises(
            ConfigurationError,
            match="rxn_params rate reaction r1 was not "
            "provided with a stoichiometry configuration "
            "argument.",
        ):
            m.rxn_params = GenericReactionParameterBlock(
                property_package=m.params,
                base_units=base_units,
                rate_reactions={"r1": {"heat_of_reaction": "foo", "rate_form": "foo"}},
            )

    @pytest.mark.unit
    def test_packed_config_dict(self, m):
        with pytest.raises(
            ConfigurationError,
            match=re.escape(
                "rxn_params[property_package] was not assigned a property package. "
                "Did you forget to unpack the configuration dictionary?"
            ),
        ):
            m.rxn_params = GenericReactionParameterBlock({"property_package": m.params})

    @pytest.mark.unit
    def test_rate_build_invalid_phase_stoichiometry(self, m):
        with pytest.raises(
            ConfigurationError,
            match="rxn_params stoichiometry for rate reaction "
            "r1 included unrecognised phase p7.",
        ):
            m.rxn_params = GenericReactionParameterBlock(
                property_package=m.params,
                base_units=base_units,
                rate_reactions={
                    "r1": {
                        "stoichiometry": {("p7", "c1"): -1, ("p1", "c2"): 2},
                        "heat_of_reaction": "foo",
                        "rate_form": "foo",
                    }
                },
            )

    @pytest.mark.unit
    def test_rate_build_invalid_component_stoichiometry(self, m):
        with pytest.raises(
            ConfigurationError,
            match="rxn_params stoichiometry for rate reaction "
            "r1 included unrecognised component c7.",
        ):
            m.rxn_params = GenericReactionParameterBlock(
                property_package=m.params,
                base_units=base_units,
                rate_reactions={
                    "r1": {
                        "stoichiometry": {("p1", "c7"): -1, ("p1", "c2"): 2},
                        "heat_of_reaction": "foo",
                        "rate_form": "foo",
                    }
                },
            )

    @pytest.mark.unit
    def test_rate_build_no_form(self, m, caplog):
        caplog.set_level(
            idaeslog.DEBUG,
            logger=("idaes.models.properties.modular_properties.base.generic_reaction"),
        )

        m.rxn_params = GenericReactionParameterBlock(
            property_package=m.params,
            base_units=base_units,
            rate_reactions={
                "r1": {
                    "stoichiometry": {("p1", "c1"): -1, ("p1", "c2"): 2},
                    "heat_of_reaction": "foo",
                }
            },
        )

        assert (
            "rxn_params rate reaction r1 was not provided with a "
            "rate_form configuration argument. This is suitable for "
            "processes using stoichiometric reactors, but not for those "
            "using unit operations which rely on reaction rate." in caplog.text
        )

    @pytest.mark.unit
    def test_equil_build(self, m):
        m.rxn_params = GenericReactionParameterBlock(
            property_package=m.params,
            base_units=base_units,
            equilibrium_reactions={
                "e1": {
                    "stoichiometry": {("p2", "c1"): -3, ("p2", "c2"): 4},
                    "heat_of_reaction": "foo",
                    "equilibrium_form": "foo",
                }
            },
        )

        rxn_config = m.rxn_params.config.equilibrium_reactions

        assert isinstance(m.rxn_params.equilibrium_reaction_idx, Set)
        assert len(m.rxn_params.equilibrium_reaction_idx) == 1
        assert "e1" in m.rxn_params.equilibrium_reaction_idx

        assert not hasattr(self, "rate_reaction_idx")

        assert isinstance(m.rxn_params.equilibrium_reaction_stoichiometry, dict)
        assert len(m.rxn_params.equilibrium_reaction_stoichiometry) == 4
        for k, v in m.rxn_params.equilibrium_reaction_stoichiometry.items():
            if (k[1], k[2]) in rxn_config[k[0]].stoichiometry.keys():
                assert v == rxn_config[k[0]].stoichiometry[k[1], k[2]]
            else:
                assert v == 0

        assert not hasattr(self, "rate_reaction_stoichiometry")

        assert isinstance(m.rxn_params.reaction_idx, Set)
        assert m.rxn_params.reaction_idx == m.rxn_params.equilibrium_reaction_idx

        assert isinstance(m.rxn_params.reaction_e1, Block)

    @pytest.mark.unit
    def test_equil_build_no_stoichiometry(self, m):
        with pytest.raises(
            ConfigurationError,
            match="rxn_params equilibrium reaction e1 was not "
            "provided with a stoichiometry configuration "
            "argument.",
        ):
            m.rxn_params = GenericReactionParameterBlock(
                property_package=m.params,
                base_units=base_units,
                equilibrium_reactions={
                    "e1": {"heat_of_reaction": "foo", "equilibrium_form": "foo"}
                },
            )

    @pytest.mark.unit
    def test_equil_build_invalid_phase_stoichiometry(self, m):
        with pytest.raises(
            ConfigurationError,
            match="rxn_params stoichiometry for equilibrium "
            "reaction e1 included unrecognised phase p7.",
        ):
            m.rxn_params = GenericReactionParameterBlock(
                property_package=m.params,
                base_units=base_units,
                equilibrium_reactions={
                    "e1": {
                        "stoichiometry": {("p7", "c1"): -3, ("p2", "c2"): 4},
                        "heat_of_reaction": "foo",
                        "equilibrium_form": "foo",
                    }
                },
            )

    @pytest.mark.unit
    def test_equil_build_invalid_component_stoichiometry(self, m):
        with pytest.raises(
            ConfigurationError,
            match="rxn_params stoichiometry for equilibrium "
            "reaction e1 included unrecognised component c7.",
        ):
            m.rxn_params = GenericReactionParameterBlock(
                property_package=m.params,
                base_units=base_units,
                equilibrium_reactions={
                    "e1": {
                        "stoichiometry": {("p2", "c7"): -3, ("p2", "c2"): 4},
                        "heat_of_reaction": "foo",
                        "equilibrium_form": "foo",
                    }
                },
            )

    @pytest.mark.unit
    def test_equil_build_no_form(self, m):
        with pytest.raises(
            ConfigurationError,
            match="rxn_params equilibrium reaction e1 was not "
            "provided with a equilibrium_form configuration "
            "argument.",
        ):
            m.rxn_params = GenericReactionParameterBlock(
                property_package=m.params,
                base_units=base_units,
                equilibrium_reactions={
                    "e1": {
                        "stoichiometry": {("p2", "c1"): -3, ("p2", "c2"): 4},
                        "heat_of_reaction": "foo",
                    }
                },
            )

    @pytest.mark.unit
    def test_rate_and_equil_build(self, m):
        m.rxn_params = GenericReactionParameterBlock(
            property_package=m.params,
            base_units=base_units,
            rate_reactions={
                "r1": {
                    "stoichiometry": {("p1", "c1"): -1, ("p1", "c2"): 2},
                    "heat_of_reaction": "foo",
                    "rate_form": "foo",
                }
            },
            equilibrium_reactions={
                "e1": {
                    "stoichiometry": {("p2", "c1"): -3, ("p2", "c2"): 4},
                    "heat_of_reaction": "foo",
                    "equilibrium_form": "foo",
                }
            },
        )

        r_rxn_config = m.rxn_params.config.rate_reactions

        assert isinstance(m.rxn_params.rate_reaction_idx, Set)
        assert len(m.rxn_params.rate_reaction_idx) == 1
        assert "r1" in m.rxn_params.rate_reaction_idx

        assert not hasattr(self, "equilibrium_reaction_idx")

        assert isinstance(m.rxn_params.rate_reaction_stoichiometry, dict)
        assert len(m.rxn_params.rate_reaction_stoichiometry) == 4
        for k, v in m.rxn_params.rate_reaction_stoichiometry.items():
            if (k[1], k[2]) in r_rxn_config[k[0]].stoichiometry.keys():
                assert v == r_rxn_config[k[0]].stoichiometry[k[1], k[2]]
            else:
                assert v == 0

        assert not hasattr(self, "equilibrium_reaction_stoichiometry")

        e_rxn_config = m.rxn_params.config.equilibrium_reactions

        assert isinstance(m.rxn_params.equilibrium_reaction_idx, Set)
        assert len(m.rxn_params.equilibrium_reaction_idx) == 1
        assert "e1" in m.rxn_params.equilibrium_reaction_idx

        assert not hasattr(self, "rate_reaction_idx")

        assert isinstance(m.rxn_params.equilibrium_reaction_stoichiometry, dict)
        assert len(m.rxn_params.equilibrium_reaction_stoichiometry) == 4
        for k, v in m.rxn_params.equilibrium_reaction_stoichiometry.items():
            if (k[1], k[2]) in e_rxn_config[k[0]].stoichiometry.keys():
                assert v == e_rxn_config[k[0]].stoichiometry[k[1], k[2]]
            else:
                assert v == 0

        assert not hasattr(self, "rate_reaction_stoichiometry")

        assert isinstance(m.rxn_params.reaction_idx, Set)
        assert m.rxn_params.reaction_idx == (
            m.rxn_params.rate_reaction_idx | m.rxn_params.equilibrium_reaction_idx
        )
        assert len(m.rxn_params.reaction_idx) == 2

        assert isinstance(m.rxn_params.reaction_r1, Block)
        assert isinstance(m.rxn_params.reaction_e1, Block)

    @pytest.mark.unit
    def test_build_parameters(self, m):
        m.rxn_params = GenericReactionParameterBlock(
            property_package=m.params,
            base_units=base_units,
            rate_reactions={
                "r1": {
                    "stoichiometry": {("p1", "c1"): -1, ("p1", "c2"): 2},
                    "heat_of_reaction": constant_dh_rxn,
                    "rate_form": "foo",
                    "parameter_data": {"dh_rxn_ref": -10000},
                }
            },
            equilibrium_reactions={
                "e1": {
                    "stoichiometry": {("p2", "c1"): -3, ("p2", "c2"): 4},
                    "heat_of_reaction": constant_dh_rxn,
                    "equilibrium_form": "foo",
                    "parameter_data": {"dh_rxn_ref": -20000},
                }
            },
        )

        assert isinstance(m.rxn_params.reaction_r1.dh_rxn_ref, Var)
        assert m.rxn_params.reaction_r1.dh_rxn_ref.fixed
        assert m.rxn_params.reaction_r1.dh_rxn_ref.value == -10000

        assert isinstance(m.rxn_params.reaction_e1.dh_rxn_ref, Var)
        assert m.rxn_params.reaction_e1.dh_rxn_ref.fixed
        assert m.rxn_params.reaction_e1.dh_rxn_ref.value == -20000


# -----------------------------------------------------------------------------
class TestGenericReactionBlock(object):
    @pytest.fixture
    def model(self):
        m = ConcreteModel()

        # Add a dummy thermo package for validation
        m.params = GenericParameterBlock(
            components={"c1": {}, "c2": {}},
            phases={
                "p1": {"equation_of_state": DummyEoS},
                "p2": {"equation_of_state": DummyEoS},
            },
            state_definition=modules[__name__],
            pressure_ref=100000.0,
            temperature_ref=300,
            base_units=base_units,
        )

        m.sblock = m.params.build_state_block([1])

        m.rxn_params = GenericReactionParameterBlock(
            property_package=m.params,
            base_units=base_units,
            rate_reactions={
                "r1": {
                    "stoichiometry": {("p1", "c1"): -1, ("p1", "c2"): 2},
                    "heat_of_reaction": constant_dh_rxn,
                    "rate_constant": arrhenius,
                    "rate_form": power_law_rate,
                    "concentration_form": ConcentrationForm.moleFraction,
                    "parameter_data": {
                        "dh_rxn_ref": -10000,
                        "arrhenius_const": 1,
                        "energy_activation": 1000,
                    },
                }
            },
            equilibrium_reactions={
                "e1": {
                    "stoichiometry": {("p2", "c1"): -3, ("p2", "c2"): 4},
                    "heat_of_reaction": constant_dh_rxn,
                    "equilibrium_constant": van_t_hoff,
                    "equilibrium_form": power_law_equil,
                    "concentration_form": ConcentrationForm.moleFraction,
                    "parameter_data": {
                        "dh_rxn_ref": -20000,
                        "k_eq_ref": 100,
                        "T_eq_ref": 350,
                    },
                },
                "e2": {
                    "stoichiometry": {("p2", "c1"): -5, ("p2", "c2"): 6},
                    "heat_of_reaction": constant_dh_rxn,
                    "equilibrium_constant": van_t_hoff,
                    "equilibrium_form": log_power_law_equil,
                    "concentration_form": ConcentrationForm.moleFraction,
                    "parameter_data": {
                        "dh_rxn_ref": -20000,
                        "k_eq_ref": 100,
                        "T_eq_ref": 350,
                    },
                },
            },
        )

        m.rblock = m.rxn_params.build_reaction_block(
            [1], state_block=m.sblock, has_equilibrium=True
        )

        return m

    @pytest.mark.unit
    def test_component_phase_lists(self, model):
        assert model.rblock[1].component_list is model.params.component_list
        assert model.rblock[1].phase_list is model.params.phase_list
        assert model.rblock[1].phase_component_set is model.params._phase_component_set

    @pytest.mark.unit
    def test_dh_rxn(self, model):
        assert isinstance(model.rxn_params.reaction_r1.dh_rxn_ref, Var)
        assert isinstance(model.rxn_params.reaction_e1.dh_rxn_ref, Var)
        assert model.rxn_params.reaction_r1.dh_rxn_ref.value == -10000
        assert model.rxn_params.reaction_e1.dh_rxn_ref.value == -20000

        assert isinstance(model.rblock[1].dh_rxn, Expression)
        assert len(model.rblock[1].dh_rxn) == 3
        assert value(model.rblock[1].dh_rxn["r1"]) == -10000
        assert value(model.rblock[1].dh_rxn["e1"]) == -20000
        assert value(model.rblock[1].dh_rxn["e2"]) == -20000

    @pytest.mark.unit
    def test_rate_constant(self, model):
        assert isinstance(model.rxn_params.reaction_r1.arrhenius_const, Var)
        assert model.rxn_params.reaction_r1.arrhenius_const.value == 1
        assert isinstance(model.rxn_params.reaction_r1.energy_activation, Var)
        assert model.rxn_params.reaction_r1.energy_activation.value == 1000

        assert isinstance(model.rblock[1].k_rxn, Expression)
        assert len(model.rblock[1].k_rxn) == 1
        assert value(model.rblock[1].k_rxn["r1"]) == value(
            1 * exp(-1000 / (constants.gas_constant * model.sblock[1].temperature))
        )

    @pytest.mark.unit
    def test_reaction_rate(self, model):
        rblk = model.rxn_params.reaction_r1
        assert isinstance(rblk.reaction_order, Var)
        assert len(rblk.reaction_order) == 4
        assert rblk.reaction_order["p1", "c1"].value == 1
        assert rblk.reaction_order["p1", "c2"].value == 0
        assert rblk.reaction_order["p2", "c1"].value == 0
        assert rblk.reaction_order["p2", "c2"].value == 0

        assert isinstance(model.rblock[1].reaction_rate, Expression)
        assert len(model.rblock[1].reaction_rate) == 1
        assert value(model.rblock[1].reaction_rate["r1"]) == value(
            model.rblock[1].k_rxn["r1"]
            * model.sblock[1].mole_frac_phase_comp["p1", "c1"] ** 1
        )

    @pytest.mark.unit
    def test_reaction_rate_None(self, model):
        model.rxn_params.config.rate_reactions.r1.rate_form = None

        with pytest.raises(
            ConfigurationError,
            match=re.escape(
                "rblock[1] Generic Reaction r1 was not "
                "provided with a rate_form configuration "
                "argument."
            ),
        ):
            model.rblock[1].reaction_rate

    @pytest.mark.unit
    def test_equilibrium_constant(self, model):
        assert isinstance(model.rxn_params.reaction_e1.k_eq_ref, Var)
        assert model.rxn_params.reaction_e1.k_eq_ref.value == 100
        assert isinstance(model.rxn_params.reaction_e1.T_eq_ref, Var)
        assert model.rxn_params.reaction_e1.T_eq_ref.value == 350

        assert isinstance(model.rblock[1].k_eq, Expression)
        assert len(model.rblock[1].k_eq) == 2
        assert value(model.rblock[1].k_eq["e1"]) == value(exp(1))
        assert value(model.rblock[1].k_eq["e2"]) == value(exp(1))

    @pytest.mark.unit
    def test_equilibrium_form(self, model):
        rblk = model.rxn_params.reaction_e1
        assert isinstance(rblk.reaction_order, Var)
        assert len(rblk.reaction_order) == 4
        assert rblk.reaction_order["p1", "c1"].value == 0
        assert rblk.reaction_order["p1", "c2"].value == 0
        assert rblk.reaction_order["p2", "c1"].value == -3
        assert rblk.reaction_order["p2", "c2"].value == 4

        rblk = model.rxn_params.reaction_e2
        assert isinstance(rblk.reaction_order, Var)
        assert len(rblk.reaction_order) == 4
        assert rblk.reaction_order["p1", "c1"].value == 0
        assert rblk.reaction_order["p1", "c2"].value == 0
        assert rblk.reaction_order["p2", "c1"].value == -5
        assert rblk.reaction_order["p2", "c2"].value == 6

        assert isinstance(model.rblock[1].equilibrium_constraint, Constraint)
        assert len(model.rblock[1].equilibrium_constraint) == 2
        assert value(model.rblock[1].equilibrium_constraint["e1"].body) == value(
            model.rblock[1].k_eq["e1"]
            - model.sblock[1].mole_frac_phase_comp["p2", "c1"] ** -3
            * model.sblock[1].mole_frac_phase_comp["p2", "c2"] ** 4
        )
        assert value(model.rblock[1].equilibrium_constraint["e2"].body) == value(
            model.rblock[1].log_k_eq["e2"]
            - model.sblock[1].log_mole_frac_phase_comp["p2", "c1"] * -5
            + model.sblock[1].log_mole_frac_phase_comp["p2", "c2"] * 6
        )

    @pytest.mark.unit
    def test_basic_scaling(self, model):
        model.rblock[1].calculate_scaling_factors()

        assert len(model.rblock[1].scaling_factor) == 7

        assert model.rblock[1].scaling_factor[model.rblock[1].dh_rxn["e1"]] == 5e-5
        assert model.rblock[1].scaling_factor[model.rblock[1].dh_rxn["r1"]] == 1e-4
        assert model.rblock[1].scaling_factor[model.rblock[1].k_eq["e1"]] == 1e-2
        assert model.rblock[1].scaling_factor[model.rblock[1].k_eq["e2"]] == 1e-2
        assert model.rblock[1].scaling_factor[model.rblock[1].log_k_eq["e1"]] == log(
            1e-2
        )
        assert model.rblock[1].scaling_factor[model.rblock[1].log_k_eq["e2"]] == log(
            1e-2
        )

        # Check that scaling factor was added to equilibrium constraint
        assert str(model.rblock[1].equilibrium_constraint["e1"].body) == str(
            (
                model.rblock[1].k_eq["e1"]
                - model.sblock[1].mole_frac_phase_comp["p2", "c1"]
                ** model.rxn_params.reaction_e1.reaction_order["p2", "c1"]
                * model.sblock[1].mole_frac_phase_comp["p2", "c2"]
                ** model.rxn_params.reaction_e1.reaction_order["p2", "c2"]
            )
            * 0.01
        )
        print(model.rblock[1].equilibrium_constraint["e2"].body)
        assert str(model.rblock[1].equilibrium_constraint["e2"].body) == str(
            (
                model.rblock[1].log_k_eq["e2"]
                - (
                    model.rxn_params.reaction_e2.reaction_order["p2", "c1"]
                    * model.sblock[1].log_mole_frac_phase_comp["p2", "c1"]
                    + model.rxn_params.reaction_e2.reaction_order["p2", "c2"]
                    * model.sblock[1].log_mole_frac_phase_comp["p2", "c2"]
                )
            )
            * log(0.01)
        )



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Test doc build.

This does *not* try to build the docs. Instead it looks
for the log of the build and looks for errors in it.
"""
# stdlib
import logging
import os
from subprocess import Popen

# third-party
import pytest


_log = logging.getLogger(__name__)


@pytest.fixture
def docs_path():
    """Find the docs.

    We have to assume that the docs are somewhere
    relative to either this file or the user's current
    working directory.
    Start with the working directory.

    Returns:
        str: Doc path or None if not found.
    """
    cwd = os.path.realpath(os.getcwd())
    # look for "./docs"
    path = os.path.join(cwd, "docs")
    if os.path.exists(path):
        return path
    # look for (last) "docs/" in full current working dir
    cwd_comp = cwd.split(os.path.sep)
    comp_cnt = len(cwd_comp)
    try:
        idx = list(reversed(cwd_comp)).index("docs")
        return os.path.sep.join(cwd_comp[: comp_cnt - idx])
    except ValueError:
        pass
    # look for docs relative to this file
    try:
        path = os.path.join(os.path.dirname(__file__), "..", "..", "docs")
        # look for {this file}/../../docs
        if os.path.exists(path):
            return path
        # look for "docs/" in file's full path
        file_comp = path.split(os.path.sep)
        comp_cnt = len(file_comp)
        try:
            idx = list(reversed(file_comp)).index("docs")
            return os.path.sep.join(file_comp[: comp_cnt - idx])
        except ValueError:
            pass
    except NameError:
        pass  # __file__ not defined(?)


ERRLOG = "sphinx-errors.txt"


@pytest.mark.unit
def test_sphinx_build_log(docs_path):
    """Check the sphinx log for errors or warnings."""
    _log.info('docs path = "{}"'.format(docs_path))
    if docs_path is None:
        _log.warning('Could not find "docs" directory')
        return
    log_path = os.path.join(docs_path, ERRLOG)
    if not os.path.exists(log_path):
        _log.warning(
            'Could not find "{}" in docs directory: {}'.format(ERRLOG, log_path)
        )
        return
    if os.stat(log_path).st_size == 0:  # file is empty - good
        return

    # Dump contents to stdout
    err_count = 0
    with open(log_path) as log:
        for line in log:
            err_count += 1
            print(line, end="")
    assert False, f"{err_count} Errors and/or Warnings found in {log_path}"


def _have_sphinx():
    """Test if a working 'sphinx-build' command exists."""
    have_sphinx = True
    try:
        Popen(["sphinx-build", "--version"]).wait()
    except:
        have_sphinx = False
    return have_sphinx


@pytest.mark.usefixtures("run_in_tmp_path")
@pytest.mark.component
def test_doctests(docs_path):
    if _have_sphinx():
        build_path = os.path.join(docs_path, "build")
        command = ["sphinx-build", "-M", "doctest", docs_path, build_path]
        proc = Popen(command)
        proc.wait(600)
        assert proc.returncode == 0



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Test no_method method to specify that no method has been assigned to calculate
a physical property for a phase.

Authors: Douglas Allan
"""

import pytest
from sys import modules

from pyomo.environ import ConcreteModel, Expression, units as pyunits

from idaes.models.properties.modular_properties.transport_properties import NoMethod

from idaes.core import declare_process_block_class, LiquidPhase, VaporPhase

from idaes.models.properties.modular_properties.base.generic_property import (
    GenericParameterData,
)
from idaes.models.properties.modular_properties.eos.ideal import Ideal
from idaes.models.properties.modular_properties.pure import ConstantProperties


@declare_process_block_class("DummyParameterBlock")
class DummyParameterData(GenericParameterData):
    def configure(self):
        self.configured = True

    def parameters(self):
        self.parameters_set = True


def define_state(b):
    b.state_defined = True


# Dummy method to avoid errors when setting metadata dict
def set_metadata(b):
    pass


def construct_dummy_model():
    m = ConcreteModel()

    m.params = DummyParameterBlock(
        components={
            "N2": {
                "visc_d_phase_comp": {"Vap": ConstantProperties},
                "parameter_data": {
                    "mw": (28.014, pyunits.g / pyunits.mol),
                },
            },
            "CO2": {
                "visc_d_phase_comp": {"Vap": ConstantProperties},
                "parameter_data": {
                    "mw": (44.009, pyunits.g / pyunits.mol),
                },
            },
        },
        phases={
            "Vap": {
                "type": VaporPhase,
                "equation_of_state": Ideal,
                "visc_d_phase": NoMethod,
            },
        },
        base_units={
            "time": pyunits.s,
            "length": pyunits.m,
            "mass": pyunits.kg,
            "amount": pyunits.mol,
            "temperature": pyunits.K,
        },
        state_definition=modules[__name__],
        pressure_ref=100000.0,
        temperature_ref=300,
    )

    m.props = m.params.state_block_class([1], defined_state=False, parameters=m.params)
    return m


@pytest.mark.unit
def test_NoMethod_visc_d_phase(caplog):
    m = construct_dummy_model()
    caplog.clear()
    NoMethod.visc_d_phase.build_parameters(m.params.Vap)
    assert len(caplog.records) == 1
    assert (
        caplog.records[0].message
        == "Skipping construction of dynamic viscosity for phase Vap"
    )
    assert NoMethod.visc_d_phase.return_expression(m.props[1], "Vap") == Expression.Skip
    assert len(caplog.records) == 1


@pytest.mark.unit
def test_NoMethod_therm_cond_phase(caplog):
    m = construct_dummy_model()
    caplog.clear()

    NoMethod.therm_cond_phase.build_parameters(m.params.Vap)
    assert len(caplog.records) == 1
    assert (
        caplog.records[0].message
        == "Skipping construction of thermal conductivity for phase Vap"
    )
    assert (
        NoMethod.therm_cond_phase.return_expression(m.props[1], "Vap")
        == Expression.Skip
    )
    assert len(caplog.records) == 1



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""Generate parameter and expression files for r134a
"""

__author__ = "John Eslick"

import os
import math
from pyomo.common.fileutils import this_file_dir
import pyomo.environ as pyo
from idaes.core.util.math import smooth_max
from idaes.models.properties.general_helmholtz.helmholtz_parameters import (
    WriteParameters,
)


def thermal_conductivity_rule(m):
    """Thermal Conductivity Rule

    Perkins, R.A.; Laesecke, A.; Howley, J.; Ramires, M.L.V.; Gurova, A.N.; Cusco, L.,
        Experimental thermal conductivity values for the IUPAC round-robin sample of
        1,1,1,2-tetrafluoroethane (R134a), NIST Interagency/Internal Report (NISTIR)
        - 6605, 2000, https://doi.org/10.6028/NIST.IR.6605.
    """

    a = [
        -1.05248e-2,
        8.00982e-5,
    ]
    b = {
        1: 1.836526,
        2: 5.126143,
        3: -1.436883,
        4: 6.261441e-1,
    }

    T = m.T_star / m.tau
    MW = m.MW / 1000.0  # kg/mol
    rho_star = m.rho_star / MW
    rho = m.delta * rho_star  # mol/m^3
    tau = m.tau
    delta = m.delta
    rho_crit = 5049.886  # mol/m^3
    lamb_red = 2.055e-3  # reducing tc W/m/K
    Tref = 561.411  # reference T K
    k = 1.380649e-23
    Pc = 4.05928  # MPa
    big_gam = 0.0496
    R0 = 1.03
    gamma = 1.239
    qd = 1892020000.0
    xi0 = 1.94e-10  # m
    nu = 0.63

    m.cp = pyo.ExternalFunction(library="", function="cp")
    m.cv = pyo.ExternalFunction(library="", function="cv")
    m.mu = pyo.ExternalFunction(library="", function="mu")
    m.itc = pyo.ExternalFunction(library="", function="itc")

    drho_dp = m.itc("r134a", delta, m.T_star / T) * rho_star * m.delta
    drho_dp_ref = m.itc("r134a", delta, m.T_star / Tref) * rho_star * m.delta

    deltchi = smooth_max(
        Pc * rho / rho_crit**2 * (drho_dp - drho_dp_ref * Tref / T), 0, eps=1e-8
    )
    xi = xi0 * (deltchi / big_gam) ** (nu / gamma)

    cp = m.cp("r134a", delta, tau)
    cv = m.cv("r134a", delta, tau)
    mu = m.mu("r134a", delta, tau) / 1e6
    y = qd * xi
    kappa_inv = cv / cp
    pi = math.pi
    Omega = 2.0 / pi * ((1.0 - kappa_inv) * pyo.atan(y) + kappa_inv * y)
    Omega0 = (
        2.0
        / pi
        * (1.0 - pyo.exp(-1.0 / (1.0 / y + 1.0 / 3.0 * (y * rho_crit / rho) ** 2)))
    )
    lambda_c = (
        1000 * MW * cp * rho * R0 * k * T / 6.0 / math.pi / mu / xi * (Omega - Omega0)
    )
    lambda_dg = a[0] + a[1] * T
    lambda_r = lamb_red * sum(bi * (rho / rho_crit) ** i for i, bi in b.items())
    return 1000 * (lambda_dg + lambda_r + lambda_c)


def viscosity_rule(m):
    """Viscosity Rule

    Perkins, R.A.; Laesecke, A.; Howley, J.; Ramires, M.L.V.; Gurova, A.N.; Cusco, L.,
        Experimental thermal conductivity values for the IUPAC round-robin sample of
        1,1,1,2-tetrafluoroethane (R134a), NIST Interagency/Internal Report (NISTIR)
        - 6605, 2000, https://doi.org/10.6028/NIST.IR.6605.
    """
    a = [
        0.355404,
        -0.464337,
        0.257353e-1,
    ]
    b = [
        -19.572881,
        219.73999,
        -1015.3226,
        2471.01251,
        -3375.1717,
        2491.6597,
        -787.26086,
        14.085455,
        -0.34664158,
    ]
    te = [
        0.0,
        -0.25,
        -0.50,
        -0.75,
        -1.00,
        -1.25,
        -1.50,
        -2.50,
        -5.50,
    ]
    c = {
        1: -20.6900719,
        2: 0.356029549,
        3: 2.11101816,
        4: 13.9601415,
        5: -4.5643502,
        6: -3.51593275,
        7: 214.76332,
        8: -0.890173375e-1,
        9: 0.100035295,
        10: 3.163695636,
    }
    T = m.T_star / m.tau
    rho = m.delta * m.rho_star / m.MW * 1000
    M = 102.031
    sigma = 0.46893
    eok = 299.363
    NA = 6.0221408e23
    Ts = T / eok
    vs = pyo.exp(sum(ai * pyo.log(Ts) ** i for i, ai in enumerate(a)))
    Bs = sum(bi * Ts**ti for bi, ti in zip(b, te))
    B = NA * sigma**3 * Bs / 1e9**3
    etas = 0.021357 * pyo.sqrt(M * T) / (sigma**2 * vs)
    tau = T / m.Tc
    delta0 = c[10] / (1 + c[8] * tau + c[9] * tau**2)
    delta = rho / 5017.053
    eta = (
        c[1] * delta
        + (c[2] / tau**6 + c[3] / tau**2 + c[4] / tau**0.5 + c[5] * tau**2) * delta**2
        + c[6] * delta**3
        + c[7] / (delta0 - delta)
        - c[7] / delta0
    )
    return etas * (1 + B * rho) + eta


def main(dry_run=False):
    """Generate parameter and expression files.

    Args:
        dry_run (bool): If dry run don't generate files

    Returns:
        None
    """
    main_param_file = os.path.join(this_file_dir(), "r134a.json")
    we = WriteParameters(parameters=main_param_file)
    we.add(
        {
            "viscosity": viscosity_rule,
            "thermal_conductivity": thermal_conductivity_rule,
        }
    )
    we.write(dry_run=dry_run)

    print("ASHRAE Offset")
    print(we.calculate_reference_offset(2.79075439914, 1.60488955608, 0, 0))
    return we


if __name__ == "__main__":
    main()
